from xml.etree import ElementTree
from typing import List
from pathlib import Path

from pybeerxml import Parser
from picobrew_server.beerxml.picobrew_recipe import PicoBrewRecipe
from picobrew_server.beerxml.picobrew_program_step import PicoBrewProgramStep


class PicoBrewRecipeParser(Parser):
    def parse(self, xml_file: Path) -> List[PicoBrewRecipe]:

        # Parse the BeerXML file
        recipes: List[PicoBrewRecipe] = super(PicoBrewRecipeParser, self).parse(
            xml_file
        )

        # Cast all recipes to PicoBrewRcipes
        recipes = [
            PicoBrewRecipe.from_beerxml_recipe(recipe, xml_file.name)
            for recipe in recipes
        ]

        # Parse the PicoBrew Program Steps
        programs = self.parse_zymatic_heating_steps(xml_file)

        # merge the parsed recipes with the PicoBrew program steps
        for (recipe, steps) in zip(recipes, programs):
            recipe.steps = steps

        return recipes

    # pylint: disable=bad-continuation
    def parse_zymatic_heating_steps(
        self, xml_file: Path
    ) -> List[List[PicoBrewProgramStep]]:
        # Extract the PicoBrew Zymatic/Z(?) specific heating/timing instructions from a recipe
        programs = []

        with xml_file.open("r") as recipe_file:
            tree = ElementTree.parse(recipe_file)

        for program_node in tree.iterfind(".//ZYMATIC"):

            steps = []
            for step_node in list(program_node):
                tag_name = self.to_lower(step_node.tag)

                if tag_name == "step":
                    step = PicoBrewProgramStep()
                    self.nodes_to_object(step_node, step)
                    steps.append(step)

            programs.append(steps)

        return programs



# System imports
from flask_cors import CORS
from flask import Flask
from picobrew_server.blueprints import frontend, picobrew_api, errors

app = Flask(__name__)
CORS(app)

app.config.update(
    SECRET_KEY="asassdfs", CORS_HEADERS="Content-Type", UPLOAD_FOLDER="recipes"
)

# ----- Routes ----------
app.register_blueprint(frontend.frontend)
app.register_blueprint(errors.errors)
app.register_blueprint(picobrew_api.picobrew_api)



from typing import Text


class PicoBrewProgramStep:
    def __init__(self):
        self.name = None
        self.temp = None
        self.time = None
        self.location = None
        self.drain = None

    def serialize(self) -> Text:
        # pylint: disable=fixme
        location_id_map = {
            "PassThrough": 0,
            "Mash": 1,
            "Adjunct1": 2,
            "Adjunct2": 3,
            "Adjunct3": 4,
            "Adjunct4": 5,
            "Pause": 6,  # TODO Verify this
        }

        # e.g. Heat to Temp,102,0,0,0
        return "{0},{1},{2},{3},{4}".format(
            self.name,
            int(self.temp),
            int(self.time),
            location_id_map[self.location],
            int(self.drain),
        )



from flask import Blueprint, render_template

errors = Blueprint("errors", __name__)

# -------- Routes --------
@errors.app_errorhandler(404)
def error_404(_error):
    return render_template("404.html"), 404


@errors.app_errorhandler(500)
def error_500(_error):
    return render_template("500.html"), 500



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['picobrew_server',
 'picobrew_server.beerxml',
 'picobrew_server.blueprints',
 'picobrew_server.utils']

package_data = \
{'': ['*'],
 'picobrew_server': ['static/css/*',
                     'static/font/material-design-icons/*',
                     'static/font/roboto/*',
                     'static/img/*',
                     'static/js/*',
                     'templates/*']}

install_requires = \
['Flask-Cors==3.0.8',
 'Flask==1.1.2',
 'Werkzeug==1.0.1',
 'pybeerxml>=1.0.8,<2.0.0',
 'webargs==6.0.0']

setup_kwargs = {
    'name': 'picobrew-server',
    'version': '1.0.1',
    'description': 'A reverse-engineered server for the Picobrew homebrewing machines',
    'long_description': '# picobrew-server\n<img src="https://img.shields.io/pypi/v/picobrew_server"> <img src="https://img.shields.io/pypi/pyversions/picobrew_server"> <img src="https://img.shields.io/github/workflow/status/hotzenklotz/picobrew-server/Test and Lint/master">\n\n\nThis project reverse-engineers a server for the proprietary PicoBrew protocol for use with the [PicoBrew Zymatic](http://www.picobrew.com/), a homebrewing machine. It is intended to provide an alternative to run the machine without a connection to the official servers at picobrew.com. Run your own server and sync your recipes offline.\n\n# HTTP API\nThe PicoBrew Zymatic\'s built-in Ardunio uses an unencrypted HTTP communication protocol. All request are `GET` requests and are not authenticated. The following documentation is based on Firmware 1.1.8.\n\n- [PicoBrew Zymatic API Docs on Postman](https://documenter.getpostman.com/view/234053/Szf54VEX?version=latest)\n- [PicoBrew Zymatic API Docs on GitHub](https://github.com/hotzenklotz/picobrew-server/wiki/PicoBrew-API)\n\n# Installation\n\n1. Install Python 3.7 or above\n2. In a terminal download, install and run the project:\n```bash\n// Download and install\npip install picobrew_server\n\n// Start the server in production mode on port 80\n\n// Windows \nset FLASK_APP=picobrew_server\nflask run --port 80 --host 0.0.0.0\n\n// OSX / Linux\nexport FLASK_APP=picobrew_server \nflask run --port 80 --host 0.0.0.0\n```\n\n- Connect the PicoBrew machine to your computer and enable DNS spoofing. Re-route `www.picobrew.com` to your computer.\n[More Details](https://github.com/hotzenklotz/picobrew-server/wiki/Install)\n\n# Development \n\n1. Install Python 3.7+ & [Poetry](https://python-poetry.org/):\n\n```bash\npip install poetry\n```\n\n2. Install all dependecies:\n\n```bash\npoetry install\n\n// Start the server on http://localhost:5000\nFLASK_APP=picobrew_server flask run\n```\n\n3. Lint, Format, and Type Check changes:\n```\npylint picobrew_server\nblack picobrew_server\nmypy picobrew_server\n```\n\n\n# Demo\nYou can try out the admin UI for uploading your XML files in this [online demo](https://picobrew.herokuapp.com). Please note, this website is for showcasing only and you should deploy your own version.\n\n\n# Features\n- Import BeerXML files\n- Send all your recipes to the PicoBrew\n- Send cleaning recipes to the PicoBrew\n- Session Logging\n- Session Recovery\n- Admin Web UI\n\nToDo\n\n- Session Charts\n\n# Machine Support\n- Picobrew Zymatic\n\nToDo\n- Picobrew Z Series\n- Picobrew Pico C\n\n# Disclaimer\nThis software is provided "as is" and any expressed or implied warranties are disclaimed. This software submits recipes with temperature targets to your PicoBrew machine and will cause it to heat water. Any damage to your PicoBrew machine is at your own risk.\n\nIf the Zymatic faults and the screen goes blank, DON\'T leave it powered on. The circulating pump will shut off and the heater stays on. A tube in the glycol loop may rupture.\n\n# License\n\nMIT @ Tom Herold\n',
    'author': 'Tom Herold',
    'author_email': 'heroldtom@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/hotzenklotz/picobrew-server',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import hashlib

from typing import Text, List
from pybeerxml import Recipe
from picobrew_server.beerxml.picobrew_program_step import PicoBrewProgramStep


def get_hash(text: Text) -> Text:
    hasher = hashlib.md5()
    hasher.update(text.encode("utf-8"))
    return hasher.hexdigest()[:32]


class PicoBrewRecipe(Recipe):
    def __init__(self, filename: Text):
        super().__init__()

        # create a unique id for every recipe based on the filename
        # pylint: disable=invalid-name
        self.id = get_hash(filename)
        self.steps: List[PicoBrewProgramStep] = []

    @classmethod
    def from_beerxml_recipe(cls, recipe: Recipe, filename: Text) -> "PicoBrewRecipe":

        picobrew_recipe = PicoBrewRecipe(filename)

        # Copy all properties of the BeerXML object
        for key, value in recipe.__dict__.items():
            picobrew_recipe.__dict__[key] = value

        return picobrew_recipe

    def serialize(self) -> Text:
        return "{0}/{1}/{2}/".format(self.name, self.id, self.get_recipe_steps(),)

    def get_recipe_steps(self) -> Text:
        steps = [step.serialize() for step in self.steps]
        return "/".join(steps)

    def save(self):
        pass



import json
import uuid
import re
import time
import logging

from pathlib import Path
from typing import Text

from flask import Blueprint, request, abort
from webargs import fields
from webargs.flaskparser import use_kwargs, parser
from picobrew_server.blueprints.frontend import get_recipes, get_recipe

from picobrew_server.utils.constants import SESSION_PATH, SYSTEM_USER

picobrew_api = Blueprint("picobrew_api", __name__)
logger = logging.getLogger()

# Unfortunately the PicoBrew API ony consist of three overloaded routes.

# ----------- RECIPES -----------
@picobrew_api.route("/API/SyncUser")
@picobrew_api.route("/API/SyncUSer")
@use_kwargs(
    {"user": fields.Str(required=True), "machine": fields.Str(required=True)},
    location="querystring",
)
def parse_recipe_request(user: Text, machine: Text):
    if user == SYSTEM_USER:
        return get_cleaning_recipes()
    if machine != "":
        return get_picobrew_recipes()

    # default response
    return "\r\n##"


@picobrew_api.route("/API/checksync")
@use_kwargs({"user": fields.Str(required=True)}, location="querystring")
# pylint: disable=unused-argument
def check_sync(user):
    return "\r\n#!#"


def get_cleaning_recipes():
    # pylint: disable=line-too-long
    return "#Cleaning v1/7f489e3740f848519558c41a036fe2cb/Heat Water,152,0,0,0/Clean Mash,152,15,1,5/Heat to Temp,152,0,0,0/Adjunct,152,3,2,1/Adjunct,152,2,3,1/Adjunct,152,2,4,1/Adjunct,152,2,5,1/Heat to Temp,207,0,0,0/Clean Mash,207,10,1,0/Clean Mash,207,2,1,0/Clean Adjunct,207,2,2,0/Chill,120,10,0,2/|Rinse v3/0160275741134b148eff90acdd5e462f/Rinse,0,2,0,5/|#"


def get_picobrew_recipes():
    all_recipes = get_recipes()

    # only send recipes with valid brewing instructions "steps" to the machine
    recipes = [recipe.serialize() for recipe in all_recipes if len(recipe.steps) > 0]
    recipes = "|".join(recipes)

    return "#{0}|#".format(recipes)


# ----------- SESSION LOGGING -----------
# pylint: disable=inconsistent-return-statements
@picobrew_api.route("/API/LogSession")
@picobrew_api.route("/API/logSession")
@picobrew_api.route("/API/logsession")
def parse_start_new_session():
    args = parser.parse(
        {
            "recipe": fields.Str(),
            "session": fields.Str(),
            "user": fields.Str(),
            "code": fields.Str(required=True),
            "firm": fields.Str(),
            "machine": fields.Str(),
            "data": fields.Str(),
            "state": fields.Str(),
            "step": fields.Str(),
        },
        request,
        location="querystring",
    )

    if "recipe" in args:
        return create_new_session(args["recipe"], args)

    if "session" in args:
        return log_to_session(args["session"], args)

    # default fallthrough
    abort(500)


def create_new_session(recipe_id: Text, _args) -> Text:
    session_id = uuid.uuid4().hex[:32]
    session_file = "{0}.json".format(session_id)

    recipe_name = None

    for recipe in get_recipes():
        if recipe.id == recipe_id:
            recipe_name = recipe.name

    session_data = {
        "date": time.strftime("%x"),
        "recipe_id": recipe_id,
        "recipe_name": recipe_name,
        "session_id": session_id,
        "steps": [],
    }

    try:
        session_directory = Path(SESSION_PATH)
        if not session_directory.exists():
            session_directory.mkdir(exist_ok=True)

        with session_directory.joinpath(session_file).open("w") as out_file:
            json.dump(session_data, out_file, indent=2)

    except IOError as error:
        logger.error("Could create new session storage file. %s", error)
        return "##"

    return "#{0}#".format(session_id)


def log_to_session(session_id: Text, args) -> Text:
    session_file = f"{session_id}.json"
    data = args["data"]
    code = int(args["code"])

    try:
        session_directory = Path(SESSION_PATH)
        if not session_directory.exists():
            session_directory.mkdir(exist_ok=True)

        with session_directory.joinpath(session_file).open("r") as in_file:
            session = json.load(in_file)

        if code == 1:
            # new program step e.g. "Heating"

            session_step = {"name": data, "temperatures": {}}
            session["steps"].append(session_step)

        elif code == 2:
            # temperature data for current step

            temps = re.findall(r"/([0-9]+)", data)
            current_time = time.strftime("%X")

            session_step = session["steps"][-1]
            session_step["temperatures"][current_time] = temps

            session["machine_state"] = args["step"]

        elif code == 3:
            # end session

            session["steps"].append({"name": "Session Ended"})

        with session_directory.joinpath(session_file).open("w") as out_file:
            json.dump(session, out_file, indent=2)

    except IOError as error:
        logging.error("Couldn't save heating log to file. %s", error)

    return ""


# ----------- SESSION RECOVERY -----------
# pylint: disable=inconsistent-return-statements
@picobrew_api.route("/API/recoversession")
@use_kwargs(
    {"session": fields.Str(required=True), "code": fields.Int(required=True)},
    location="querystring",
)
def parse_session_recovery_request(session, code) -> Text:
    try:
        session_file = f"{session}.json"

        with Path(SESSION_PATH).joinpath(session_file).open("r") as in_file:
            session = json.load(in_file)

        if code == 0:
            # return a recipe
            try:
                recipe_file = Path("recipes").joinpath(session["recipe_filename"])
                recipes = get_recipe(recipe_file)
                recipe = recipes[
                    0
                ]  # per spec a BeerXML file can contain more than one recipe
                return f"#{recipe.serialize()}|!#"

            except IOError as error:
                logging.error(
                    "Unable to resume session %s. Recipe %s. %s",
                    error,
                    recipe_file,
                    session,
                )
                abort(500)

        elif code == 1:
            # return machine params
            return f"#{session['machine_state']}#"

    except IOError as error:
        logging.error(
            "Unable to resume session %s. Recipe %s. %s", error, recipe_file, session
        )
        abort(500)

    # default fallthrough
    abort(500)



import os
import logging
from pathlib import Path
from typing import Text, List

from flask import Blueprint, request, flash, redirect, url_for, render_template, session
from werkzeug.utils import secure_filename
from picobrew_server.beerxml.picobrew_parser import PicoBrewRecipeParser, PicoBrewRecipe

from picobrew_server.utils.constants import ALLOWED_FILE_EXTENSIONS

logger = logging.getLogger()
frontend = Blueprint("frontend", __name__)

# -------- Routes --------
@frontend.route("/")
def index():
    return render_template("index.html")


@frontend.route("/recipes")
def render_recipes():
    return render_template("recipes.html", recipes=get_recipes())


def get_recipes(recipe_path: Text = "recipes"):

    files = [
        filename
        for filename in Path(recipe_path).glob("**/*")
        if filename.suffix in ALLOWED_FILE_EXTENSIONS
    ]

    recipes = [get_recipe(filename) for filename in files]
    recipes = [y for x in recipes for y in x]  # flatten

    return recipes


def get_recipe(filename: Path) -> List[PicoBrewRecipe]:
    try:
        parser = PicoBrewRecipeParser()
        return parser.parse(filename)

    # pylint: disable=broad-except
    except Exception as error:
        logger.error("Failed to parse recipe %s. %s", filename, error)
        return []


@frontend.route("/upload", methods=["POST"])
def upload_recipe():

    redirect_url = ".index"
    for file in request.files.getlist("recipes"):

        file_directory = Path("recipes")
        file_directory.mkdir(exist_ok=True)

        filename = file_directory.joinpath(secure_filename(file.filename))

        if filename.suffix in ALLOWED_FILE_EXTENSIONS:
            file.save(filename)
            redirect_url = ".validate"
            session["recipe_file"] = str(filename)
        else:
            flash("Invalid BeerXML file <%s>." % file.filename)

    return redirect(url_for(redirect_url))


@frontend.route("/validate")
def validate():

    filename = Path(session["recipe_file"])
    recipe = get_recipe(filename)[0]
    return render_template("validate.html", recipe=recipe)


@frontend.route("/submit_eula", methods=["POST"])
def submit_eula():

    redirect_url = ".render_recipes"
    form_data = request.form

    if not form_data.getlist("accept_eula") or form_data.getlist("action") == "cancel":
        file = session["recipe_file"]
        os.remove(file)
        redirect_url = ".index"

    return redirect(url_for(redirect_url))


# -------- Template Utility --------
@frontend.context_processor
def utility_processor():
    def format_weight(amount, _unit="kg"):
        if amount < 1.0:
            format_string = "{0:.0f}{1}".format(amount * 1000, "g")
        else:
            format_string = "{0:.2f}{1}".format(amount, "kg")
        return format_string

    def format_time(time):
        if time < 24 * 60:
            format_string = "{0:.0f}{1}".format(time, "min")
        else:
            format_string = "{0:.0f}{1}".format(time / (24 * 60), "days")
        return format_string

    def format_volume(volume, unit="L"):
        return "{0:.2f}{1}".format(volume, unit)

    def format_float(value, trailing_numbers):
        return "{0:.{1}f}".format(value, trailing_numbers)

    return dict(
        format_weight=format_weight,
        format_time=format_time,
        format_volume=format_volume,
        format_float=format_float,
    )
